/* vim: set sw=8 ts=8 si : */
/* Author: Guido Socher, Copyright: Guido Socher, BSD style license, the same licence conditions as easyexif, see easyexif/LICENSE.txt */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#ifdef DO_CHMOD
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include "easyexif/exif.h"
// the version string is shown as part of the help text (starting with version 1.1)
#define VERSION_STR "1.5"
// the exif tag should be at the beginning of the image file, typically in the first 100Kbytes
#define BUFSIZE 300000
unsigned char *buf = new unsigned char[BUFSIZE];
static char opt_c='\0';
#ifdef PROMPT_TO_CLOSE
static int opt_n=0;
#endif
static int opt_p=0;

// this function uses the global variable buf
// dstr is the date string and it is an out-value
int get_exifdate(char *filename,std::string &dstr){
	unsigned long bytesread=0;
        unsigned long fsize;
	FILE *fp = fopen(filename, "rb");
	if (!fp) { 
		printf("IGNORE: can not open file %s\n",filename); 
		return(0); 
	}
	// Read the start of JPG file into a buffer
        fseek(fp,0,SEEK_END);
	fsize=ftell(fp);
	rewind(fp);
	if (fsize>BUFSIZE){
		fsize=BUFSIZE;
	}
	bytesread=fread(buf, 1, fsize, fp);
	fclose(fp);
	if (bytesread!=fsize){
		printf("IGNORE: can't read file  %s\n",filename);
		return(0); 
	}
	// Parse EXIF
	EXIFInfo result;
	int code = result.parseFrom(buf, bytesread);
	if (code) {
		if (code==PARSE_EXIF_ERROR_NO_JPEG){
			printf("IGNORE: data is not a jpg file, %s\n",filename);
		}else if(code==PARSE_EXIF_ERROR_NO_EXIF){
			printf("IGNORE: no exif information found, %s\n",filename);
		}else if(code==PARSE_EXIF_ERROR_UNKNOWN_BYTEALIGN){
			printf("IGNORE: unknown byte alignment, %s\n",filename);
		}else if(code==PARSE_EXIF_ERROR_CORRUPT){
			printf("IGNORE: exif data corrupted, %s\n",filename);
		}else{
			printf("IGNORE: Error parsing EXIF: code %d, %s\n", code,filename);
		}
		return(0);
	}
	// Dump EXIF information for debug purposes
	/*
	printf("Camera make       : %s\n", result.Make.c_str());
	printf("Camera model      : %s\n", result.Model.c_str());
	printf("Image width       : %d\n", result.ImageWidth);
	printf("Image height      : %d\n", result.ImageHeight);
	printf("Original date/time: %s\n", result.DateTimeOriginal.c_str());
	printf("Image date/time   : %s\n", result.DateTime.c_str());
	printf("Digitize date/time: %s\n", result.DateTimeDigitized.c_str());
	*/
	if (result.DateTimeOriginal.length() > 15){
		dstr.assign(result.DateTimeOriginal);
	}else if(result.DateTimeDigitized.length() > 15){
		dstr.assign(result.DateTimeDigitized);
	}else if(result.DateTime.length() > 15){
		dstr.assign(result.DateTime);
	}else{
		printf("IGNORE: no date information for %s\n",filename);
		return(0);
	}
	// the time stamps look like this:
	// 2014:04:06 11:39:33
	// check that it is really a string starting with 19.. or 20..
	if (((dstr[0]=='1' && dstr[1]=='9')||(dstr[0]=='2' && dstr[1]=='0'))&&
	     dstr[4]==':' && dstr[7]==':' && dstr[10]==' ' && dstr[13]==':'){
		dstr[4]='-';
		dstr[7]='-';
		dstr[10]='_';
		dstr.erase(16,dstr.length()-16); // cut at the end
		dstr.erase(13,1); // remove : between hour and min
	}else{
		printf("IGNORE: unsupported exif date format %s, %s\n",dstr.c_str(),filename);
		return(0);
	}
	return(1);
}

int ends_in_jpg(const char *s){
	int l;
	l=strlen(s);
	if (l<4) return(0);
	if (strcasecmp(&(s[l-4]),".jpg")==0) return(1);
	if (l<5) return(0);
	if (strcasecmp(&(s[l-5]),".jpeg")==0) return(1);
	return(0);
}

// find the last occurance of a given char in string and return the position
// to it (first char=pos 0)
int findchar(const char *s, const char c){
	int l,i,pos;
	l=strlen(s);
	i=0;
	pos=0;
	while(i<l){
		if (s[i]==c){
			pos=i;
		}
		i++;
	}
	return(pos);
}

void help(){
	printf("exifdate -- rename .jpg using the exif date information inside the image\n");
#ifdef PROMPT_TO_CLOSE
	printf("USAGE: exifdate [-h] [-c C][-n][-p] filename1.jpg ...\n\n");
#else
	printf("USAGE: exifdate [-h] [-c C][-p] filename1.jpg ...\n\n");
#endif
	printf("exifdate can be used to apply consistent names to images taken with\n");
	printf("different cameras. It is useful for sorting  and archiving digital images.\n");
	printf("The file names are constructed such that they sort automatically in time\n");
	printf("order in your file manager.  The date/time information is read out of\n");
	printf("the image exif data.\n");
	printf("\nOPTIONS:\n");
	printf("-c letter to add to serial number. This can be used to categorize images.\n");
	printf("   The file name will normally contain a 3 digit serial number and this\n");
	printf("   option sets the left most position to a fixed character.\n");
	printf("   example: exifdate -c j PICT0301.jpg\n");
	printf("            PICT0301.jpg-> 2014-05-20_0722-j01.jpg\n");
	printf("-h this help\n");
#ifdef PROMPT_TO_CLOSE
	printf("-n terminate program without \"press enter to continue...\"\n");
#endif
	printf("-p just print timestamps, do not change anything\n");
	printf("\nThis is exifdate version: %s\n",VERSION_STR);
	exit(0);
}

int main(int argc, char *argv[]) {
	int charpos;
	int serial=0;
	char serialnumstr[20];
	char newfilename[400];
	std::string timestamp;
	// The following things are used for getopt: 
        extern char *optarg __attribute__ ((unused));
        extern int optind;
        extern int opterr;
	int ch;

	opterr=0; // prevent getopt's own error handling, go to '?'
	if (argc<2) help();
	while ((ch = getopt(argc, argv, "c:hnp")) != -1) {
		switch (ch) {
		case 'c':
			opt_c=optarg[0];
			break;
		case 'h':
			help(); // will exit the code
		case 'n':
#ifdef PROMPT_TO_CLOSE
			opt_n=1;
#else
			fprintf(stderr, "ERROR: No such option. -h for help.\n");
			exit(1);
#endif
			break;
		case 'p':
			opt_p=1;
			break;
		case '?':
			fprintf(stderr, "ERROR: No such option. -h for help.\n");
			exit(1);
		//no default action for case 
		}
	}
	if (optind == argc){
		help();
	}
	while(optind<argc){
		if (!ends_in_jpg(argv[optind])){
			printf("IGNORE: not .jpg or .jpeg, %s\n",argv[optind]);
			optind++;
			continue;
		}
		if (get_exifdate(argv[optind],timestamp)){
			//printf("timestamp %s\n",timestamp.c_str());
			charpos=findchar(argv[optind],'.'); // the . before jpg
			// extract the last 3 serial number digits
			// that the camera puts into the file names
			// and use it if available:
			if (charpos>3 && isdigit(argv[optind][charpos-1]) && isdigit(argv[optind][charpos-2]) && isdigit(argv[optind][charpos-3])){
				strncpy(serialnumstr,&(argv[optind][charpos-3]),3);
				serialnumstr[3]='\0';
			}else{
				// fallback in case there is no serial number from the camera:
				serial++;
				sprintf(serialnumstr,"%03d",serial);
			}
			if (opt_c){
				// we have a character to put into the serial number string:
				serialnumstr[0]=opt_c;
			}
			charpos=findchar(argv[optind],'/'); // the last /
			if (charpos==0){
				// perhaps windows
				charpos=findchar(argv[optind],'\\');
			}
			if (charpos>0){
				// step after the "/" or "\" if we have one (convert to length of the string to copy)
				charpos++;
			}
			if (charpos+10+timestamp.length() < sizeof(newfilename)){
				newfilename[0]='\0';
				if (charpos>0){
					strncpy(newfilename,argv[optind],charpos);
					newfilename[charpos]='\0';
				}
				strcat(newfilename,timestamp.c_str());
				strcat(newfilename,"-");
				strcat(newfilename,serialnumstr);
				strcat(newfilename,".jpg");
				if (opt_p){
					// print in a different way so we know it is not renaming:
					printf("%s : %s\n",argv[optind],newfilename);
					// just print and continue
					optind++;
					continue;
				}else{
					printf("%s-> %s\n",argv[optind],newfilename);
				}
				if (access(newfilename, F_OK )==0){
					printf("ERROR: file %s already exists\n",newfilename);
					optind++;
					continue;
				}
				if(rename(argv[optind],newfilename)!=0){
					perror("rename error");
#ifdef DO_CHMOD
				}else{
					chmod(newfilename,0644);
#endif
				}
			}else{
				printf("ERROR: path too long, %s\n",argv[optind]);
			}
		}
		optind++;
	}
	// under the windows graphical environment a dos-shell will
	// automatically be opened but it closes so fast that one can not
	// see the text. 
#ifdef PROMPT_TO_CLOSE
	if (opt_n==0){
		printf("OK: please PRESS ENTER to continue...\n");
		fgetc(stdin);
	}
#endif
	delete[] buf;
	return (0);
}
// =====================
